-- onEvent
local function onEvent(self, event, a, b, c)
	self[event](self, a, b, c)
end

-- doUpdate
local function Update(self)

	local myZone = GetRealZoneText()
	local list = {}

	for unitid, unitName, unitClass, group, zone, online, dead in XPerl_NextMember do
		if (self.group[group]) then
			if (not self.sameZone or (zone == myZone)) then
				tinsert(list, {["group"] = group, name = unitName})
			end
		end
	end

	sort(list, function(a,b) return a.group..a.name < b.group..b.name end)

	local text = ""
	local totals = 0
	for k,v in pairs(list) do
		text = text..v.name.."\r"
		totals = totals + 1
	end

	XPerl_RosterText.text = text
	self.textFrame.scroll.text:SetText(text)
	self.textFrame.scroll.text:HighlightText()
	self.textFrame.scroll.text:SetFocus()

	self.totals:SetText(format(XPERL_ROSTERTEXT_TOTAL, totals))
end

-- scalingUpdate
function XPerl_RosterText_ScalingUpdate(self)
	if (XPerl_RosterText) then
		local w = XPerl_RosterTextAnchor:GetWidth()
		if (w) then
			-- h 350, w 250

			local s = min(5, max(0.4, w / 350))

			XPerl_RosterTextAnchor:SetWidth(350 * s)
			XPerl_RosterTextAnchor:SetHeight(250 * s)

			XPerl_RosterText:SetScale(s)
		end
	end
end

-- StartScaling(self)
local function StartScaling(self)
	XPerl_RosterTextAnchor:StartSizing()
	self.textFrame.scroll.text:HighlightText(0, 0)
end

local function StopScaling(self)
	XPerl_RosterTextAnchor:StopMovingOrSizing()
	self.textFrame.scroll.text:HighlightText()
end

-- XPerl_RosterText_Init
function XPerl_RosterText_Init(self)

	self.group = {1, 1, 1, 1, 1, nil, nil, nil}
	self.sameZone = nil

	self:SetScript("OnEvent", onEvent)
	self.Update = Update
	self.StartScaling = StartScaling
	self.StopScaling = StopScaling

	self.RAID_ROSTER_UPDATE = Update
	self.PLAYER_ENTERING_WORLD = Update

	self:SetScript("OnShow", function(self)
		self:RegisterEvent("RAID_ROSTER_UPDATE")
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		Update(self)
	end)
	self:SetScript("OnHide", function(self)
		XPerl_RosterText.text = nil
		self.textFrame.scroll.text:SetText("")
		self:UnregisterAllEvents()
	end)

	XPerl_RosterText_Init = nil
end
